#pragma once
enum PixelDir { RIGHT, LEFT };
enum PixelWhen { PREVIOUS, CURRENT };

extern Adafruit_NeoPixel pixels;

//--------- NeoPixel Definitions
struct PixelState {       /* stores the current and the previous NeoPixel state for both pixels */
	int red;
	int green;
	int blue;
} pixelState[2][2];

/* setPixelColor: sets the specified Pixel to its color and saves the previous state */
void setPixelColor(enum PixelDir pixelDir, int red, int green, int blue) {
	struct PixelState now = { red, green, blue };

	pixels.setPixelColor(pixelDir, red, green, blue);
	pixels.show();

	memcpy(&pixelState[pixelDir][PREVIOUS], &pixelState[pixelDir][CURRENT], sizeof(pixelState[pixelDir][CURRENT]));
	memcpy(&pixelState[pixelDir][CURRENT], &now, sizeof(now));
}

/* setPreviousPixelColor: restores the previous pixel color for the specified pixel and sets it as the current one */
void setPreviousPixelColor(enum PixelDir pixelDir) {
	struct PixelState flip;

	pixels.setPixelColor(pixelDir, pixelState[pixelDir][PREVIOUS].red,
		pixelState[pixelDir][PREVIOUS].green,
		pixelState[pixelDir][PREVIOUS].blue);
	pixels.show();

	memcpy(&flip, &pixelState[pixelDir][CURRENT], sizeof(pixelState[pixelDir][CURRENT]));
	memcpy(&pixelState[pixelDir][CURRENT], &pixelState[pixelDir][PREVIOUS], sizeof(pixelState[pixelDir][PREVIOUS]));
	memcpy(&pixelState[pixelDir][PREVIOUS], &flip, sizeof(flip));
}