#include <gfxfont.h>
#include <Adafruit_GFX.h>
#include <Wire.h>
#include <Encoder.h>
#include <Adafruit_NeoPixel.h>
#include <Ticker.h>
#include <ESP8266WiFi.h>
#include <PubSubClient.h>
#include "MutichannelGasSensor.h"
#include <Adafruit_GFX.h>
#include <Adafruit_IS31FL3731.h>
#include "Classifier.h"
#include "LinkedList.h"
#include "WlanGraphics.h"
#include "NeoPixels.h"

#define WIFI_SSID             "enose-net"
#define WIFI_PW               "enose-pw-UCB"
#define MQTT_SERVER_ADDRESS   "thingspeak.umwelt-campus.de"
#define MQTT_SERVER_PORT      1883

// turn functionality on or off
// klassengenerierung zur laufzeit
#define F_NEW_FEATURES          0
// mqtt zu node-red
#define F_ONLINE_MQTT           0
// plott
#define F_SERIAL_PLOTTER        1
//mit gelbem lernmodus
#define F_LEARNMODE             0
 


#define numFeatures 3
#define k 3
#define memSize 30
#define prevEntries 3
//delay in einem interrupt max 16383 button press
#define DELAY_MYS_MAX 16383

enum beverages {  SPRUDEL, APFELSAFT, WEIN, NUMLABELS };
String labels[]{ "Sprudel", "Apfelsaft", "Wein" }; 
String labelconcat = "";
int numClasses = NUMLABELS;
//sensor nach messung für 30 sec gesperrt
volatile bool isLocked = false;

//gedaechtnis
const int previousX[prevEntries][numFeatures] = {
  {968, 1000, 338},
  {971,  989, 364},
  {963,   58, 317}
};

enum beverages previousY[prevEntries] = {
  SPRUDEL, APFELSAFT, WEIN
};

Classifier<int> *knn;


// Encoder-Library http://www.pjrc.com/teensy/td_libs_Encoder.html
Encoder button_encoder(14,12); // Button rotation encoder object
WiFiClient   espClient; 
PubSubClient mqttclient(espClient);
Adafruit_NeoPixel pixels = Adafruit_NeoPixel(2,13,NEO_GRBW + NEO_KHZ800);

//--------- CharlieWing Matrix Display Definitions
Ticker matrixOutput;
Ticker lockTicker;
Ticker checkForSerialInputTicker;
volatile int matrixOutput_index = 0; /* current position in matrix */
String matrixOutput_text  = " ";     /* output string as global variable */
volatile int matrixOutput_wait  = 0; /* to wait until the (scrolling) text has run through */
volatile bool matrixOutput_changed = false;
volatile bool matrixForceOut = false;
byte measurementId = 0;

Adafruit_IS31FL3731_Wing matrix = Adafruit_IS31FL3731_Wing();


volatile bool button_pressed = false;
volatile bool mode_switch_pressed = false;

enum channelSelect {_NH3 = 1, _CO, _NO2};

//------------ reconnect mqtt-client
void mqttreconnect() { // Loop until we're reconnected 
  if (!mqttclient.connected()) 
    while (!mqttclient.connected()) 
      mqttclient.connect("online-learner-nose");
  else { 
    mqttclient.loop(); 
  }
}

/* matrixUpdate: updates the matrix output */
void matrixUpdate(){
  int numberPixels=(matrixOutput_text.length())*6;
  if (numberPixels > 15) { /* scrolling */
    if (matrixOutput_index >= -numberPixels) {
      matrixOutput_changed;
      matrix.clear();
      matrix.setCursor(matrixOutput_index + 5,0);
      matrix.print(matrixOutput_text);
      matrixOutput_index--;
    } else {
      matrix.clear();
      matrixOutput_index = 0;
      matrixOutput_wait=0;
    }
  } else { /* only three symbols max, no scrolling */
    matrix.setCursor(0,0);
    matrix.print(matrixOutput_text);
    matrixOutput_wait  = 0; /* no waiting */
  }
}

/* matrixDisplay: sets the text to output */
void matrixDisplay(String text, int wait) {
  if (matrixOutput_text  != text) { /* update only when text is different from before */
    matrix.clear();
    matrixOutput_index = 0;
    matrixOutput_text  = text;
    matrixOutput_wait  = wait;
    while (matrixOutput_wait && !matrixForceOut) delay(10); /* wait until text is completely displayed once */
  } 
};

void connectPC(){
  int received = 0;
  int connd = 0;
  String inputString = "";
  String decision;
  
    while (Serial.available() && ! connd) {
      received = 1;
      // get the new byte:
      char inChar = (char)Serial.read();
      // add it to the inputString:
      inputString += inChar;
  
      if (received){
        decision = inputString.substring(0, 2);
        if (decision.equals("Is")){
            Serial.println(String(String("Yes:") + labelconcat).c_str());
            connd = 1;
        }
      }
    }
  }



//--------- button definitions
/* buttonPress_ISR: Interrupt service routine executed on button press */
void buttonPress_ISR() {
   int i;
  /* button debouncing */
   int old_key = digitalRead(2);
   int new_key;
   delayMicroseconds(10000); 
   new_key = digitalRead(2);
    if (new_key != old_key){
      return;
   }

   for (i = 0; i < 25; i++)
    delayMicroseconds(DELAY_MYS_MAX);

  if (digitalRead(2)==LOW) /* button was held down to switch execution modes */
      mode_switch_pressed = true;
  else
      button_pressed = true;
}
   

void unlock(){
  isLocked = false;
  setPixelColor(LEFT,0,0,0);
  }

void send_label_if_not_yet() {
  static bool label_sent = false;
  if (WiFi.status() == WL_CONNECTED && !label_sent) {
    mqttreconnect();
    mqttclient.publish("e-nose/labels", labelconcat.c_str(), true); //retained
    label_sent = true;
  }
}

void setup()
{
  // set labelconcat right from the start, as it is needed in serial interrupt
  int numberOfClasses = sizeof(labels) / sizeof(*labels);
  for (int i = 0; i < numberOfClasses - 1; i++)
    labelconcat += (labels[i] + ",");
  labelconcat += labels[numberOfClasses - 1];

    Serial.begin(115200);  // start serial for output
    gas.begin(0x04);//the default I2C address of the slave is 0x04
    gas.powerOn();

    pixels.begin(); /* initlalize NeoPixels */
    delay(1);

    setPixelColor(RIGHT,0,0,0); /* turn NeoPixels off */
    setPixelColor(LEFT,0,0,0);

    matrix.begin(); /* initialize CharlieWing Display Matrix */
    delay(10);
    matrix.clear(); 
    matrix.setTextColor(60); /* limit brightness */ 
    matrixOutput.attach(0.08, matrixUpdate); /* cyclic update */

#if F_SERIAL_PLOTTER
    checkForSerialInputTicker.attach(1, connectPC);
#endif

    attachInterrupt(2, buttonPress_ISR, FALLING);

    knn = new Classifier<int>(numClasses, k, memSize, numFeatures, prevEntries, (int *)previousX, (size_t *)previousY);
  // knn->print_memory();
    setPixelColor(RIGHT,40,0,0); // red

#if F_ONLINE_MQTT    
    // connecting to WLAN
    WiFi.persistent(false);
    WiFi.mode(WIFI_STA);
    delay(100);

    
    WiFi.begin(WIFI_SSID,WIFI_PW);
    int wlan_imagecounter = 0;
    while (WiFi.status() != WL_CONNECTED && !button_pressed) { // Warte bis Verbindung steht 
      switch (wlan_imagecounter) {
        case 0: matrix.clear(); matrix.drawBitmap(0,0,onebar_bmp,15,7,60); break;
        case 1: matrix.drawBitmap(0,0,twobars_bmp,15,7,60); break;
        case 2: matrix.drawBitmap(0,0,threebars_bmp,15,7,60); break;
        }
        ++wlan_imagecounter %= 3;
        
      delay(300);
    };

    if (button_pressed){
      /* aborting wlan connection, show abort sign */
      matrix.clear(); 
      matrix.drawBitmap(0,0,abort_bmp,15,7,60);
      delay(1000);
      button_pressed = false;
      }

    mqttclient.setServer(MQTT_SERVER_ADDRESS, MQTT_SERVER_PORT);
    
    // send label via mqtt
  send_label_if_not_yet();
#endif
    
    // preheating 10 minutes
    matrix.clear();
    for (int i = 0; i < 600; i++){
      if (button_pressed)
        // preheating aborted
        break;
        
      matrix.drawFastVLine(i/40, 0, 7, (i % 40)*3/4);
          
      delay(1000);
      
    }
    
    matrix.clear(); 

    
}

void loop()
{

  static enum modes {CLASSIFYMODE, LEARNMODE} curMode = CLASSIFYMODE;
  static enum states {READY, MEASURING, CLASS_READY, LEARNSTATE, TEST_MEASUREMENT} state = READY;
  static int curTrainingClass, rs0, rs1, rs2;
  switch (state){
    // state for testing purposes
    case TEST_MEASUREMENT:
    {
      static int timer = 0;
      int rs0 = gas.get_addr_dta(_NH3);
      int rs1 = gas.get_addr_dta(_CO);
      int rs2 = gas.get_addr_dta(_NO2);

      Serial.print(rs0),Serial.print(","),Serial.print(rs1),Serial.print(","),
      Serial.print(rs2),Serial.print(","),Serial.print(curTrainingClass),Serial.print(","),Serial.println(timer++);
      delay(1000);
      if (button_pressed){
        button_pressed = false;
        timer = 0;
        }
      break;
    }
    case READY:
    {
      setPixelColor(RIGHT,40,40,40); // white
      button_pressed = false;
      mode_switch_pressed = false;
      matrixDisplay("", 0);
      while ((!button_pressed || isLocked) && !mode_switch_pressed){
        if (isLocked)
          button_pressed = false;
        delay(10);
      }

      if (mode_switch_pressed){
        mode_switch_pressed = false;
        if (F_LEARNMODE) {
          state = LEARNSTATE;
          curMode = LEARNMODE;
          }
        break;
        }
      state = MEASURING;
      break;
    }
    case MEASURING:
    {
      matrixDisplay("", 0);
      setPixelColor(LEFT,40,0,40); // purple
      delay(5000); // 5 seconds
      setPixelColor(LEFT,0,40,0); // green

      rs0 = gas.get_addr_dta(_NH3);
      rs1 = gas.get_addr_dta(_CO);
      rs2 = gas.get_addr_dta(_NO2);

#if F_SERIAL_PLOTTER      
      Serial.print(rs0);
      Serial.print(",");
      Serial.print(rs1);
      Serial.print(",");
      Serial.println(rs2);
#endif

      setPixelColor(LEFT,20,0,0); // red
      isLocked = true;
      lockTicker.once(50,unlock);

      switch (curMode){
        case CLASSIFYMODE:
        {
          int data[] = {rs0, rs1, rs2};
          size_t *resultvec = new size_t[numClasses];
          int result = knn->classify(data, resultvec);
         // knn->print_memory();
          size_t confidence = resultvec[result];

#if F_ONLINE_MQTT
          if (WiFi.status() == WL_CONNECTED){
        send_label_if_not_yet();
              // send via mqtt
              byte mqttmsg[] = {++measurementId,
                                rs0 >> 24, rs0 >> 16, rs0 >> 8, rs0,
                                rs1 >> 24, rs1 >> 16, rs1 >> 8, rs1,
                                rs2 >> 24, rs2 >> 16, rs2 >> 8, rs2,
                                (byte)result};
    
              mqttreconnect();
              mqttclient.publish("e-nose/classified",mqttmsg, (byte)14);
          }
#endif

          int countDraw = 0;
          for (int i = 0; i < numClasses; i++)
            if (resultvec[i] == confidence)
              countDraw++;

      delete[] resultvec;
          
          String confString = "";
          if (countDraw > 1)
            confString = "?";
          else if (confidence > k-0.1)
            confString = "!";
            
      if (result < NUMLABELS)
        matrixDisplay(labels[result] + confString + String("  ") + labels[result] + confString, 1);
      else
      {
        matrixDisplay(String(result+1, DEC) + confString, 1);
        delay(5000);
      }
          matrixDisplay("",0);
          
          state = CLASS_READY;
          break;
        }
        case LEARNMODE:
        {
          int data[] = {rs0, rs1, rs2};
          knn->learn(data, (size_t) curTrainingClass-1);

#if F_ONLINE_MQTT
         if (WiFi.status() == WL_CONNECTED){
       send_label_if_not_yet();
          // send via mqtt
          byte mqttmsg[] = {++measurementId, 
                            rs0 >> 24, rs0 >> 16, rs0 >> 8, rs0,
                            rs1 >> 24, rs1 >> 16, rs1 >> 8, rs1,
                            rs2 >> 24, rs2 >> 16, rs2 >> 8, rs2,
                            (byte)curTrainingClass-1};

          mqttreconnect();
          mqttclient.publish("e-nose/learned",mqttmsg, (byte)14);
          }
#endif

#if F_SERIAL_PLOTTER
          Serial.print("learn:");
          Serial.println(curTrainingClass-1);
#endif
          
         // knn->print_memory();
          state = LEARNSTATE;
          break; // switch curMode
        }
      }
      break; // switch state
    }

    case CLASS_READY:
    {
      setPixelColor(RIGHT,40,40,40); // white
      int32_t buttonOffset = button_encoder.read();
      float trainingPosition;
      float buttonPos;
      int newOffset = 1;  // 1 = no new features ; 2 = new features
      button_pressed = false;
      mode_switch_pressed = false;
      int newTrainingClass;

#if F_NEW_FEATURES
      newOffset = 2;
#endif      
      do {
          buttonPos = (button_encoder.read() - buttonOffset) % 96;
          trainingPosition = ((buttonPos < 0) ? (buttonPos + 96) : (buttonPos)) / 96.0;
          
          if (trainingPosition > (0.5/(numClasses+newOffset)) && trainingPosition < 1-(0.5/(numClasses+newOffset))){
              newTrainingClass = ((trainingPosition + (0.5 /(numClasses + newOffset))) * (numClasses + newOffset));

              if (curTrainingClass != newTrainingClass){
              curTrainingClass = newTrainingClass;
 
        String curLabel;

        if (curTrainingClass <= NUMLABELS)
          curLabel = labels[curTrainingClass - 1];  // class has a label
        else if (curTrainingClass <= numClasses)
          curLabel = String(curTrainingClass, DEC); // class exists but has no label
        else
          curLabel = "Neu";             // new class field


              matrixForceOut = true;
              matrixDisplay(curLabel, 0);
              matrixForceOut = false;
              }
          }
          else
            if (curTrainingClass != -1){
            curTrainingClass = -1;
            matrixDisplay(String(""), 0);
            }
          if (isLocked)
              button_pressed = false;
          delay(10);
      } while ((!button_pressed) && !mode_switch_pressed);

      if(mode_switch_pressed){
          state = LEARNSTATE;
          curMode = LEARNMODE;
          break;
      }
      if (curTrainingClass == -1){
        if (!isLocked)
          state = MEASURING;
        break;
      }
      else
      {
      if (curTrainingClass > numClasses)    // make new class
      {
        knn->add_class();
        numClasses++;
      }

          int data[] = {rs0, rs1, rs2};
          knn->learn(data, (size_t) curTrainingClass-1);
#if F_ONLINE_MQTT
          if (WiFi.status() == WL_CONNECTED){
        send_label_if_not_yet();
              // send via mqtt
              byte mqttmsg[] = {measurementId,
                                rs0 >> 24, rs0 >> 16, rs0 >> 8, rs0,
                                rs1 >> 24, rs1 >> 16, rs1 >> 8, rs1,
                                rs2 >> 24, rs2 >> 16, rs2 >> 8, rs2,
                                (byte)curTrainingClass-1};
    
              mqttreconnect();
              mqttclient.publish("e-nose/learned",mqttmsg, (byte)14);
          }
#endif
#if F_SERIAL_PLOTTER
          Serial.print("learn:");
          Serial.println(curTrainingClass-1);
#endif
        //  knn->print_memory();
          matrixDisplay(String(""), 0);
          state = READY;
          break;
      }
      
    }
    case LEARNSTATE:
    {
      setPixelColor(RIGHT,40,40,0); // yellow
      float trainingPosition;
      float buttonPos;
      bool firstRun = true;
      button_pressed = false;
      mode_switch_pressed = false;
      int newTrainingClass;
      int numDisplayLabels = numClasses;
      
#if F_NEW_FEATURES
      numDisplayLabels++;
#endif      
      do {
          buttonPos = (button_encoder.read()) % 96;
          trainingPosition = ((buttonPos < 0) ? (buttonPos + 96) : (buttonPos)) / 96.0;
          

          newTrainingClass = (trainingPosition + (0.5 / numDisplayLabels))* numDisplayLabels;
          newTrainingClass -= floor(newTrainingClass/ numDisplayLabels) * numDisplayLabels;
          newTrainingClass++;
  
            if (curTrainingClass != newTrainingClass || firstRun){
                  firstRun = false;
                  curTrainingClass = newTrainingClass;
              String curLabel;
      
              if (curTrainingClass <= NUMLABELS)
                curLabel = labels[curTrainingClass - 1];  // class has a label
              else if (curTrainingClass <= numClasses)
                curLabel = String(curTrainingClass, DEC); // class exists but has no label
              else
                curLabel = "Neu";             // new class field
      
                      matrixForceOut = true;
                      matrixDisplay(curLabel, 0);
                      matrixForceOut = false;
             }
          delay(10);
          if (isLocked)
            button_pressed = false;
      } while ((!button_pressed || isLocked) && !mode_switch_pressed);

      if(mode_switch_pressed){
          mode_switch_pressed = false;
          curMode = CLASSIFYMODE;
          state = READY;
          break;
      }

    if (curTrainingClass > numClasses)    // make new class
    {
      knn->add_class();
      numClasses++;
    }

          state = MEASURING;
          break;
    }
  }

}
